function demo_sog_correlate(type)
% This demo shows the properties of my Sum of Gaussian (SoG) scan matching algorithm.
% There are two algorithms, described in Sections 4.4.4 (Algorithm 4.2) and Section 4.4.5 of my thesis:
%
%	T. Bailey. Mobile Robot Localisation and Mapping in Extensive Outdoor Environments. 
%   PhD thesis, University of Sydney, Australian Centre for Field Robotics, 2002.
%
% The first is the basic correlation of two SoGs, and the second is variant that considers
% the observation SoG as a set of independent measurements. I also provide a MatLab implementation of 
% the first algorithm to show that it gives the same numbers as the mex (C++) version. However, be warned 
% that the MatLab version is *very* slow.
%
% This demo computes the correlation of two scans along slices of the state-space (ie, the relative pose)
% Each slice shows the shape of the likelihood function along an orthogonal axis passing through the nominal
% pose value. A simple hill-climbing algorithm moves the nominal pose estimate towards the maximum likelihood.
%
% A note about algorithm 1. From the graphs in this demo it can be seen that the maximum likelihood value
% of this correlation method is very accurate and reliable. However, the variance of the correlation graph
% is *very* over-conservative. This is due to the underlying observation model as a "single point target" as
% discussed in Sections 4.2 and 4.3 of my thesis.
%
% A note about algorithm 2. This algorithm produces a much less conservative correlation graph, but is
% sensitive to outliers, viewpoint variation etc. It requires an accurate initial estimate of the relative pose
% and appropriate tuning of a "probability of false measurement" parameter.
%
% Tim Bailey 2004.

format compact
load laserdata % logged laser range scans from indoor robot
figure, title('Correlation slices')

% Configurable values
knn = 20; % number of near-neighbours (0 means all neighbours)
PF = 0.1; % probability false detection (sog_correlate_variant only)
ign = 0;  % ignore outliers (sog_correlate_variant only)

% Constants
MAXR = 7.8;                     % this laser data times-out at ~7.9 m
R = [0.05^2 0; 0 (pi/360)^2];   % range-bearing covariance matrix
phi = -pi/2:(pi/360):pi/2;      % scan bearing angles

% Convert 2 scans to SoGs
r1 = laser(554, 2:end);
r2 = laser(556, 2:end);
Sr = make_gaussian_sum(r1, phi, R, MAXR); % reference scan
So = make_gaussian_sum(r2, phi, R, MAXR); % observation scan

% Compute correlation along a slice, and hill-climb to find best
rel = [0;0;0];    % initial relative pose estimate
srch = -1:0.02:1; % search slice 
for i=1:6
    x = rel(1) + srch;
    y = rel(2) + srch;
    phi = rel(3) + srch*5*pi/180;
    
    % compute correlation for each slice
    tic
    for j=1:length(srch)
        if type == 1 % demo mex implementation
            wx(j) = sog_correlate(Sr, So, [x(j); rel(2:3)], knn);
            wy(j) = sog_correlate(Sr, So, [rel(1); y(j); rel(3)], knn);
            wp(j) = sog_correlate(Sr, So, [rel(1:2); phi(j)], knn);
            
        elseif type == 2 % demo matlab implementation (very slow)
            wx(j) = sog_correlate_m(Sr, So, [x(j); rel(2:3)]);
            wy(j) = sog_correlate_m(Sr, So, [rel(1); y(j); rel(3)]);
            wp(j) = sog_correlate_m(Sr, So, [rel(1:2); phi(j)]);
            toc
            
        elseif type == 3 % demo variance correlation (mex)
            wx(j) = sog_correlate_variant(Sr, So, [x(j); rel(2:3)], knn, PF, ign);
            wy(j) = sog_correlate_variant(Sr, So, [rel(1); y(j); rel(3)], knn, PF, ign);
            wp(j) = sog_correlate_variant(Sr, So, [rel(1:2); phi(j)], knn, PF, ign);

        else
            error('Invalid type: must be 1 to 3')
        end
    end
    toc
    
    % plot results
    subplot(3,1,1), hx = plot(x, wx); xlabel(sprintf('x-axis, given y=%f and phi=%f',rel(2),rel(3))), ylabel('correlation'), title('Correlation slices')
    subplot(3,1,2), hy = plot(y, wy); xlabel(sprintf('y-axis, given x=%f and phi=%f',rel(1),rel(3))), ylabel('correlation')
    subplot(3,1,3), hp = plot(phi, wp); xlabel(sprintf('phi-axis, given x=%f and y=%f',rel(1),rel(2))), ylabel('correlation')
    drawnow
    
    % hill-climb (note, this is a demo of sog_correlate, *not* optimisation)
    ix = find(wx == max(wx));
    iy = find(wy == max(wy));
    ip = find(wp == max(wp));    
    if wx(ix) >= wy(iy) & wx(ix) >= wp(ip)
        rel(1) = x(ix);
    elseif wy(iy) >= wp(ip)
        rel(2) = y(iy);
    else
        rel(3) = phi(ip);
    end
end