function w = sog_correlate_m(Sr, So, x)
%
% Inputs:
%   Sr, So - Sums of Gaussians
%   x - relative pose [x;y;phi] of So with respect to Sr
%
% Output:
%   w - correlation weight for given pose
%
% Notes: 
%
% 1. The SoGs are represented by arrays of structures, where each struct has 
% fields {x, P, w}. These are the mean [x;y], variance and weight of each Gaussian.
%
% 2. This algorithm is given in Section 4.4.4 (Algorithm 4.2) of my thesis:
%
%	T. Bailey. Mobile Robot Localisation and Mapping in Extensive Outdoor Environments. 
%   PhD thesis, University of Sydney, Australian Centre for Field Robotics, 2002.
%
% Tim Bailey 2004.

c = cos(x(3)); 
s = sin(x(3));
R = [c -s; s c];

xog = transform_to_global([So.x], x);

w = 0;
for i=1:length(So)
    xi = xog(:,i);
    Pi = R * So(i).P * R';
    wi = So(i).w;

    for j=1:length(Sr)
        v = xi - Sr(j).x;
        S = Pi + Sr(j).P;
        w = w + wi * Sr(j).w * compute_gauss(v,S);
    end    
end

%
%

function w = compute_gauss(v,S)
% compute value of 2-D Gaussian
denom = 2*pi * sqrt(det(S));
num = exp(-0.5 * v' * inv(S) * v);
w = num/denom;
