/*--------------------
 
	[x,a,b] = icp(a, b, x0, gate, n, interp)

	INPUTS:
	a,b - two sets of points p = [x1, x2, ...; y1, y2, ...]
	x0 - initial guess at relative pose
	gate - maximum distance check
	n - number of iterations
	interp - perform interpolation 

	OUTPUT:
	x - pose of b relative to a [x; y; p]
	a,b - associated points used to compute x

	Tim Bailey 2004.

/*--------------------*/

#include "mex.h"
#include "../source/icp.hpp"
#include <cassert>

using namespace std;
using namespace Geom2D;

#define PNTA	prhs[0]
#define PNTB	prhs[1]
#define POSE0	prhs[2]
#define GATE	prhs[3]
#define NITS	prhs[4]
#define INTERP	prhs[5]
#define POSE	plhs[0]
#define ASSOCA	plhs[1]
#define ASSOCB	plhs[2]

#define MINPNTS	3

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[] )     
{
	// check arguments
    if (nrhs != 6 || nlhs>3) 
		mexErrMsgTxt("Usage: [x,a,b] = icp(a, b, x0, gate, n, interp)"); 
    if (mxGetM(PNTA)!=2 || mxGetM(PNTB)!=2 || mxGetM(POSE0)!=3)
		mexErrMsgTxt("Incorrect input matrix sizes.");
	if (mxGetN(PNTA) < MINPNTS || mxGetN(PNTB) < MINPNTS)
		mexErrMsgTxt("Insufficient data points.");

	// arrange data
	Point* pa= reinterpret_cast<Point*>(mxGetPr(PNTA)); 
	Point* pb= reinterpret_cast<Point*>(mxGetPr(PNTB)); 
	vector<Point> va(pa, pa+mxGetN(PNTA));
	vector<Point> vb(pb, pb+mxGetN(PNTB));

	Pose init= *reinterpret_cast<Pose*>(mxGetPr(POSE0));
	double gate = *mxGetPr(GATE);
	int nits = static_cast<int>(*mxGetPr(NITS));
	bool interp = static_cast<bool>(*mxGetPr(INTERP));

	// create pose output
	POSE = mxCreateDoubleMatrix(3, 1, mxREAL);
	Pose* pse = reinterpret_cast<Pose*>(mxGetPr(POSE)); 

	// algorithm
	ICP icp(va, gate);
	*pse = icp.align(vb, init, gate, nits, interp);
	const std::vector<Point>& aa = icp.get_points_ref();
	const std::vector<Point>& ab = icp.get_points_obs();
	assert(aa.size() == ab.size());

	// create association output
	ASSOCA = mxCreateDoubleMatrix(2, aa.size(), mxREAL);
	ASSOCB = mxCreateDoubleMatrix(2, aa.size(), mxREAL);
	Point* paa = reinterpret_cast<Point*>(mxGetPr(ASSOCA)); 
	Point* pab = reinterpret_cast<Point*>(mxGetPr(ASSOCB)); 
	for (int i = 0; i < aa.size(); ++i) {
		paa[i] = aa[i];
		pab[i] = ab[i];
	}
}
