/*--------------------
 
	nn = k_neighbours(a,b, k, gate)

	INPUTS:
	a,b - two sets of points [x; y]
	k - for each b, find k near-neighbours in a
	gate - maximum permitted distance to a NN

	OUTPUTS:
	nn - array of structures (one for each b)
	has:
		nn(i).ai - indices to k neighbours in a
		nn(i).d - distance to each ai from b(i)

/*--------------------*/

#include "mex.h"
#include "../source/nn.hpp"

using namespace std;
using namespace Geom2D;

void store_neighbours(mxArray *out, int idx, 
					  const std::vector<int>& ni, const std::vector<double>& nd)
{
	const int size = ni.size();
	mxArray *ai = mxCreateDoubleMatrix(1,size,mxREAL); 
	mxArray *ad = mxCreateDoubleMatrix(1,size,mxREAL);

	double *pi = mxGetPr(ai);
	double *pd = mxGetPr(ad);
	for (int i=0; i<size; ++i) {
		*pi++ = ni[i]+1;
		*pd++ = nd[i];
	}

	mxSetFieldByNumber(out, idx, 0, ai);
	mxSetFieldByNumber(out, idx, 1, ad);
}

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
	// check arguments
    if (nrhs != 4 || nlhs > 1) 
		mexErrMsgTxt("Usage: nn = k_neighbours(a, b, k, gate)"); 
    if (mxGetM(prhs[0])!=2 || mxGetM(prhs[1])!=2)
		mexErrMsgTxt("Incorrect input matrix sizes.");

	// sizes
	int na = mxGetN(prhs[0]);
	int nb = mxGetN(prhs[1]);
	int k = static_cast<int>(*mxGetPr(prhs[2]));
	if (na < k)
		k = na;

	// get input parameters
	Point* p= reinterpret_cast<Point*>(mxGetPr(prhs[0])); 
	vector<Point> pa(p, p + na);
	Point* pb= reinterpret_cast<Point*>(mxGetPr(prhs[1]));
	double gate = static_cast<double>(*mxGetPr(prhs[3]));

	// create output structure array
	const char *fields[] = { "ai", "d" };
	const int nfields = sizeof(fields)/sizeof(fields[0]);
	plhs[0] = mxCreateStructArray(1, &nb, nfields, fields);

	// algorithm
	SweepSearch nn(pa, gate);

	std::vector<int> idx(k);
	for (int i=0; i<nb; ++i) {
		const std::vector<double>& nd = nn.query(pb[i], idx);
		store_neighbours(plhs[0], i, idx, nd);
	}
}
