/*=================================================================
 *
 * [x, w] = sog_optimise(Sr, So, x, nits, k)
 *
 * INPUTS:
 *	Sr, So - Sums of Gaussians: represented by array of structures
 *		with fields {x, P, w}. These are the mean [x;y], variance and 
 *		weight of each Gaussian.
 *	x - relative pose of So with respect to Sr.
 *	nits - number of optimisation iterations.
 *	k - if > 0, means compute likelihood using k-nearest-neighbours.
 *
 * OUTPUT:
 *	x - optimal relative pose
 *  w - optimal correlation weight.
 *
 * Tim Bailey 2004.
 *
 *=================================================================*/

#include "mex.h"
#include "../source/sog_match.hpp"

using namespace std;
using namespace Geom2D;

void get_SoG_data(std::vector<SoG>& s, const mxArray *in)
{
	const int size = mxGetNumberOfElements(in);
	s.resize(size);

	for (int i=0; i<size; ++i) {
		mxArray *x = mxGetFieldByNumber(in, i, 0);
		mxArray *P = mxGetFieldByNumber(in, i, 1);
		mxArray *w = mxGetFieldByNumber(in, i, 2);
		double *pP = mxGetPr(P);

		s[i].x = *reinterpret_cast<Point *>(mxGetPr(x));
		s[i].P = Matrix(pP[0], pP[2], pP[1], pP[3]);
		s[i].w = *mxGetPr(w);
	}
}

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
	// check arguments
	if (nrhs != 5 || nlhs > 2)
		mexErrMsgTxt("Usage: [x, w] = sog_optimise(Sr, So, x, nits, k)"); 
	if (mxIsStruct(prhs[0]) == 0 ||
		mxIsStruct(prhs[1]) == 0)
		mexErrMsgTxt("Error: Sr and So must be structures."); 
	if (mxGetM(prhs[2]) != 3)
		mexErrMsgTxt("Error: x must be 3 x 1 array."); 

	// extract input data
	vector<SoG> Sr;
	vector<SoG> So;
	get_SoG_data(Sr, prhs[0]);
	get_SoG_data(So, prhs[1]);
	Pose rel = *reinterpret_cast<Pose *>(mxGetPr(prhs[2]));
	int nits = static_cast<int>(*mxGetPr(prhs[3]));
	int k = static_cast<int>(*mxGetPr(prhs[4]));

	// algorithm
	sog_optimise sopt(Sr, So, (k>0)?k:0); 
	double w = sopt.optimise_pose(rel, nits);

	// create output
	plhs[0] = mxCreateDoubleMatrix(3, 1, mxREAL); 
	plhs[1] = mxCreateDoubleMatrix(1, 1, mxREAL); 
	*reinterpret_cast<Pose *>(mxGetPr(plhs[0])) = rel;
	*mxGetPr(plhs[1]) = w;
}
