/* Compute likelihood function for the cross-correlation of two Gaussians.
 * The basic algorithm is given in Section 4.4.4 (Algorithm 4.2) of my thesis:
 *
 *	T. Bailey. Mobile Robot Localisation and Mapping in Extensive Outdoor
 *	Environments. PhD thesis, University of Sydney, Australian Centre for
 *	Field Robotics, 2002.
 *
 * Tim Bailey 2004.
 */
#ifndef LIKELIHOOD_FUNC_H_
#define LIKELIHOOD_FUNC_H_

#include "geometry2D.hpp"
#include <cmath>

namespace Geom2D {

//
// 2D Matrix operations
//

struct Matrix { // 2x2 Matrix
	Matrix() {}
	Matrix(double a, double b,
		   double c, double d) : xx(a), xy(b), yx(c), yy(d) {}
	double xx, xy;
	double yx, yy;
};

inline Point operator- (const Point& p, const Point& q)
{
	Point r(p);
	r.x -= q.x;
	r.y -= q.y;
	return r;
}

inline Point operator+ (const Point& p, const Point& q)
{
	Point r(p);
	r.x += q.x;
	r.y += q.y;
	return r;
}

inline Matrix operator+ (const Matrix& m, const Matrix& n)
{
	Matrix r(m);
	r.xx += n.xx; r.xy += n.xy;
	r.yx += n.yx; r.yy += n.yy;
	return r;
}

inline Point operator* (const Matrix& m, const Point& p)
{
	Point r;
	r.x = m.xx*p.x + m.xy*p.y;
	r.y = m.yx*p.x + m.yy*p.y;
	return r;
}

inline Matrix operator* (const Matrix& m, const Matrix& n)
{
	Matrix r;
	r.xx = m.xx*n.xx + m.xy*n.yx; r.xy = m.xx*n.xy + m.xy*n.yy;
	r.yx = m.yx*n.xx + m.yy*n.yx; r.yy = m.yx*n.xy + m.yy*n.yy;
	return r;
}

inline double determinant(const Matrix &m)
{
	return m.xx*m.yy - m.xy*m.yx;
}

inline Matrix inverse(const Matrix &m)
{
	double det = determinant(m);
	Matrix t(m.yy/det, -m.xy/det,
			-m.yx/det,  m.xx/det);
	return t;
}

inline Matrix transpose(const Matrix &m)
{
	Matrix t(m);
	t.xy = m.yx;
	t.yx = m.xy;
	return t;
}

inline Matrix rotation_matrix(const double& angle) 
// Compute a matrix for rotating 2D points by angle radians.
{
	using namespace std;
	return Matrix(cos(angle), -sin(angle),
				  sin(angle),  cos(angle));
}

//
// Likelihood function operations
//

inline double scale_volume(const Matrix &P)
// Calculate the volume of Gaussian G(0,P), if the Gaussian peak is scaled to 1.
{
	using namespace std;
	return 2.*PI*sqrt(determinant(P));
}

#if 1

inline double calculate_w(const Point &v, const Matrix &S) 
// Compute the density of a Gaussian G(0,S) at offset v.
{
	using namespace std;
	double det = determinant(S);
	double vSv = ((v.x*S.yy - v.y*S.yx)*v.x + (v.y*S.xx - v.x*S.xy)*v.y) / det;
	return exp(-0.5*vSv) / (2.*PI*sqrt(det));
}

#else

inline double calculate_w(const Point &v, const Matrix &S) 
// Compute the density of a Gaussian G(0,S) at offset v.
// TODO: requires overloaded operators to be defined.
{
	using namespace std;
	double det = determinant(S);
	Matrix Si = inverse(S);
	double vSv = v*Si*v;
	return exp(-0.5*vSv) / (2.*PI*sqrt(det));
}

#endif

inline double likelihood(const Point& p0, const Matrix& P0, 
				  const Point& p1, const Matrix& P1, const Pose& rel)
// Compute the likelihood function for the correlation of two Gaussians.
// Inputs: Gaussians G(p0,P0), G(p1,P1), and their relative pose.
// Output: likelihood weight.
{
	// Convert G(p1,P1) to the coordinate-frame of G(p0,P0)
	Matrix R = rotation_matrix(rel.phi); 
	Point p1g = rel.p + R*p1;
	Matrix P1g = R*P1*transpose(R);

	Point v = p0-p1g; // innovation
	Matrix S= P0+P1g;
	return calculate_w(v,S);
}

} // namespace Geom2D 

#endif
