/////////////////////////////////////////////
//
// Downhill simplex (Nelder-Mead) algorithm. Adapted from Numerical Recipes
// in C, 1992.
#include "simplex.hpp"

simplex::simplex(const val_type& start, const val_type& lamda, const double& ftol,
	const simplex_functor& f) : func(f), tol(ftol), num_its(0)
{
	v_sum.resize(start.size());
	f_vals.resize(start.size()+1);
	vertices.resize(start.size()+1);

	make_vertices(start, lamda);
	vertex_sum();
	set_best();
}

void simplex::reset(const val_type& lamda)
{
	make_vertices(vertices[ilo], lamda);
	vertex_sum();
	set_best();
}

bool simplex::optimise(int num)
{
	for (int i=0; i<num; ++i) {
		get_vertex_indices();
		
		double f_try=func_eval(-1.0);

		if (f_try<=f_vals[ilo])
			f_try=func_eval(2.0);

		else if (f_try >= f_vals[ihi2]) {
			double f_old=f_vals[ihi2];
			f_try=func_eval(0.5);

			if (f_try >= f_old) 
				contract();
		}
	}
	
	set_best();
	return f_vals[ihi]-f_vals[ilo] < tol;
}

void simplex::make_vertices(const val_type& start, const val_type& lamda)
{
	vertices[0]=start;
	f_vals[0]=func(vertices[0]);

	for (int i=0, j=1; i<lamda.size(); ++i, ++j) {
		vertices[j]=vertices[0];
		vertices[j][i]+=lamda[i];
		f_vals[j]=func(vertices[j]);
	}

	num_its+=vertices.size();
}

void simplex::vertex_sum() 
{
	for (int i=0; i<v_sum.size(); ++i) {
		double sum=0.0;
		for (int j=0; j<vertices.size(); ++j)
			sum += vertices[j][i];
		v_sum[i]=sum;
	}
}

void simplex::get_vertex_indices()
{
	if (f_vals[0]<f_vals[1]) {
		ilo=ihi2=0;
		ihi=1;
	}
	else {
		ilo=ihi2=1;
		ihi=0;
	}

	for (int i=2; i<f_vals.size(); ++i) {
		if (f_vals[i]<f_vals[ilo])
			ilo=i;
		else if (f_vals[ihi]<f_vals[i]) {
			ihi2=ihi;
			ihi=i;
		}
	}
}

double simplex::func_eval(double factor) 
{
	val_type v_try(v_sum.size());
	double f1=(1.0-factor)/v_sum.size();
	double f2=f1-factor;

	for (int i=0; i<v_sum.size(); ++i)
		v_try[i] = v_sum[i]*f1 - vertices[ihi][i]*f2;

	double f_try = func(v_try);
	++num_its;

	if (f_try < f_vals[ihi]) {
		f_vals[ihi]=f_try;

		for(int i=0; i<v_sum.size(); ++i) {
			v_sum[i] += v_try[i]-vertices[ihi][i];
			vertices[ihi][i] = v_try[i];
		}
	}

	return f_try;
}

void simplex::contract()
{
	for (int i=0; i<vertices.size(); ++i)
		if (i!=ilo) {
			for (int j=0; j<v_sum.size(); ++j) {
				v_sum[j]=0.5*(vertices[i][j]+vertices[ilo][j]);
				vertices[i][j]=v_sum[j];
			}
			f_vals[i]=func(v_sum);
			++num_its;
		}

	vertex_sum();
}

void simplex::set_best()
{
	get_vertex_indices();
	f_best=f_vals[ilo];
	v_best=vertices[ilo];
}
