////////////////////////////
//

#include "sog_match.hpp"
#include "simplex.hpp"
#include <algorithm>
#include <cassert>

using namespace std;

namespace Geom2D {

class MaxPose : public simplex_functor {
public:
	MaxPose(const sog_correlate& scn) : sc(scn) {}
	double operator() (const val_type& x) const {
		Pose p = {x[0], x[1], x[2]};
		return -sc.get_weighting(p);
	}

private:
	const sog_correlate& sc;
};

sog_correlate::sog_correlate(const vector<SoG>& Sr_, const vector<SoG>& So_, 
							 int k, double pf, bool ignore)
: Sr(Sr_), So(So_), idx(k), nn(0), Pf(pf), IgnoreOutlier(ignore)
{
	assert(k < Sr.size());

	if (k > 0) { // for nearest-neighbour correlate
		pr.reserve(Sr.size());
		for (int i=0; i<Sr.size(); ++i)
			pr.push_back(Sr[i].x);

		nn = new SweepSearch(pr,9e9); // TODO: 9e9 is a hack for 'large distance'
	}
	else { // for full correlate
		idx.reserve(Sr.size());
		for (int i=0; i<Sr.size(); ++i)
			idx.push_back(i);
	}
}

double sog_correlate::get_weighting(const Pose& rel) const
// Weighting for SoG correlation given relative pose, rel (see thesis Section 4.4.4).
{
	double s = sin(rel.phi);
	double c = cos(rel.phi);
	Matrix R(c, -s, s, c);

	Transform2D tr(rel);

	double w=0.;
	for (int i=0; i<So.size(); ++i) {

		Point xo = So[i].x; // transform i-th obs-Gaussian to ref-frame
		tr.transform_to_global(xo);
		Matrix Po = R * So[i].P * transpose(R);
		double wo = So[i].w;

		if (nn) // knn search
			(void) nn->query(xo, idx);

		for (int j=0; j < idx.size(); ++j) { // compute likelihoods
			const SoG& SR = Sr[idx[j]];
			Point v = xo - SR.x;
			Matrix S = Po + SR.P;

			w += wo * SR.w * calculate_w(v, S);
		}
	}

	return w;
}

double sog_correlate::get_weighting_variant(const Pose& rel) const
// Alternative weighting (see my thesis Section 4.4.5)
{
	double s = sin(rel.phi);
	double c = cos(rel.phi);
	Matrix R(c, -s, s, c);

	Transform2D tr(rel);

	if (wo.size() != So.size())
		wo.resize(So.size());

	double w = log(1.);
	for (int i=0; i<So.size(); ++i) {

		Point xo = So[i].x; // transform i-th obs-Gaussian to ref-frame
		tr.transform_to_global(xo);
		Matrix Po = R * So[i].P * transpose(R);

		if (nn) // knn search
			(void) nn->query(xo, idx);

		double wi = 0;
		for (int j=0; j < idx.size(); ++j) { // compute likelihood for i-th obs
			const SoG& SR = Sr[idx[j]];
			Point v = xo - SR.x;
			Matrix S = Po + SR.P;

			wi += SR.w * calculate_w(v, S);
		}
		//wi *= So[i].w; // TODO: if we scale by So[i].w, what are we doing??
		wo[i] = wi; // store likelihood

		// Apply probability "false alarm". Here we use one of two strategies:
		//	(i) Set wi to minimum "false alarm" probability.
		//	(ii) Set wi to 1, effectively ignoring the outlier observation.
		if (wi < Pf) 
			wi = (IgnoreOutlier) ? 1. : Pf; 

		w += log(wi); // sum log-likelihoods (for numerical stability)
	}

	return exp(w);
}

sog_optimise::sog_optimise(const vector<SoG>& Sr, const vector<SoG>& So, int k)
: sc(Sr,So,k)
{
}

double sog_optimise::optimise_pose(Pose &p, int num_its) const
// simplex optimisation 
{
	assert(num_its > 0);
	const double FTOL(0.001);
	const Pose LAMDA = {0.1, 0.1, 0.01};
	
	val_type start(3); start[0]=p.p.x; start[1]=p.p.y; start[2]=p.phi;
	val_type lamda(3); lamda[0]=LAMDA.p.x; lamda[1]=LAMDA.p.y; lamda[2]=LAMDA.phi;

	MaxPose mp(sc);
	simplex s(start, lamda, FTOL, mp);

	s.optimise(num_its);

	const val_type& x=s.optimal_state();
	p.p.x = x[0]; p.p.y = x[1]; p.phi = x[2];
	return -s.optimal_value();
}
 
} // namespace Geom2D 
